"""
Plots and movies
----------------
"""

from __future__ import annotations

import multiprocessing
import subprocess
import time
from os import linesep
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FFMpegWriter
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.image import AxesImage

from pigreads.progress import PROGRESS_ITERS
from pigreads.schema.simulation import Simulation


def imshow_defaults(
    array: np.ndarray[Any, Any] | None = None,
    sim: Simulation | None = None,
    dx: float | None = None,
    dy: float | None = None,
    Nx: int | None = None,  # pylint: disable=invalid-name
    Ny: int | None = None,  # pylint: disable=invalid-name
    **kwargs: Any,
) -> dict[str, Any]:
    """
    Set default imshow arguments.

    :param array: Array to display.
    :param sim: Simulation object.
    :param dx: Grid spacing in x.
    :param dy: Grid spacing in y.
    :param Nx: Number of grid points in x.
    :param Ny: Number of grid points in y.
    :param kwargs: Additional arguments.
    :return: Dictionary of arguments for
             :py:func:`matplotlib.pyplot.imshow`.
    """

    if "origin" not in kwargs:
        kwargs["origin"] = "lower"

    if "interpolation" not in kwargs:
        kwargs["interpolation"] = "none"

    if array is not None:
        if "vmin" not in kwargs:
            kwargs["vmin"] = np.nanmin(array)

        if "vmax" not in kwargs:
            kwargs["vmax"] = np.nanmax(array)

    if sim is not None:
        dx, dy = sim.dx, sim.dy
        Nx, Ny = sim.Nx, sim.Ny

    if array is not None:
        Ny, Nx = array.shape[-2:]

    if (
        "extent" not in kwargs
        and dx is not None
        and dy is not None
        and Nx is not None
        and Ny is not None
    ):
        kwargs["extent"] = (
            -0.5 * dx,
            (Nx + 0.5) * dx,
            -0.5 * dy,
            (Ny + 0.5) * dy,
        )

    return kwargs


def plot_frame(
    ax: Axes,
    frame: np.ndarray[Any, Any],
    xlabel: str = "x",
    ylabel: str = "y",
    vlabel: str = "",
    title: str = "",
    **kwargs: Any,
) -> tuple[AxesImage, Colorbar]:
    """
    Display a frame as an image.

    :param ax: Axes object.
    :param xlabel: Label for the x-axis.
    :param ylabel: Label for the y-axis.
    :param vlabel: Colorbar label.
    :param title: Title of the plot.
    :param frame: Frame to display.
    :param kwargs: Passed to :py:func:`matplotlib.pyplot.imshow`.

    :return: Image and colorbar objects.
    """
    assert frame.ndim == 2
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    im = ax.imshow(frame, **imshow_defaults(array=frame, **kwargs))
    cbar = plt.colorbar(im)
    cbar.set_label(vlabel)
    return im, cbar


def movie(
    path: str,
    frames: np.ndarray[Any, Any],
    dpi: int = 180,
    fps: int = 15,
    tlables: list[str] | None = None,
    progress: str = "none",
    progress_dict: dict[str, int] | None = None,
    parallel: int = 1,
    **kwargs: Any,
) -> None:
    """
    Render a chunk of frames to a movie file, with optional parallelization.

    :param path: Path to write the movie file to.
    :param frames: Array of frames.
    :param dpi: Dots per inch.
    :param fps: Frames per second.
    :param tlables: List of time labels.
    :param progress: Progress bar type.
    :param progress_dict: Dictionary to store progress.
    :param parallel: Number of processes (default 1, 0 to use all CPUs).
    :param kwargs: Passed to :py:func:`plot_frame`.
    """
    assert frames.ndim == 3

    if parallel == 1:
        fig, ax = plt.subplots(dpi=dpi)
        writer = FFMpegWriter(fps=fps)
        tlables = [f"{i}" for i, _ in enumerate(frames)] if tlables is None else tlables
        im, _ = plot_frame(ax, frames[0], **imshow_defaults(array=frames, **kwargs))

        prog = PROGRESS_ITERS[progress]
        with writer.saving(fig, path, fig.dpi):
            for (i, frame), tlabel in zip(
                enumerate(prog(frames)), tlables, strict=True
            ):
                ax.set_title(tlabel)
                im.set_data(frame)
                writer.grab_frame()
                if progress_dict is not None:
                    progress_dict[path] = i + 1
        return

    Np = multiprocessing.cpu_count() if parallel == 0 else parallel  # pylint: disable=invalid-name
    Nfr = len(frames)  # pylint: disable=invalid-name
    Nfrp = (Nfr + Np - 1) // Np  # pylint: disable=invalid-name
    chunks = [slice(i, min(i + Nfrp, Nfr)) for i in range(0, Nfr, Nfrp)]

    with TemporaryDirectory() as temp:
        paths = [str(Path(temp) / f"{i}.mp4") for i, _ in enumerate(chunks)]

        with multiprocessing.Manager() as manager:
            assert progress_dict is None
            progress_dict_ = manager.dict(dict.fromkeys(paths, 0))
            tasks = [
                [
                    {
                        "path": path,
                        "frames": frames[chunk],
                        "tlables": tlables[chunk] if tlables else None,
                        "progress_dict": progress_dict_,
                        "parallel": 1,
                        "dpi": dpi,
                        "fps": fps,
                        **kwargs,
                    }
                ]
                for path, chunk in zip(paths, chunks, strict=True)
                if chunk.start < chunk.stop
            ]

            progress_proc = multiprocessing.Process(
                target=movie_progress_updater,
                args=(progress, Nfr, progress_dict_),
            )
            progress_proc.start()

            with multiprocessing.Pool(processes=Np) as pool:
                pool.starmap_async(movie_wrapper, tasks).get()

            progress_proc.join()

        pathlist = Path(temp) / "files.txt"
        with pathlist.open("w") as f:
            for p in paths:
                f.write(f"file '{p}'{linesep}")

        proc = subprocess.Popen(  # pylint: disable=consider-using-with
            [
                "ffmpeg",
                "-y",
                "-f",
                "concat",
                "-safe",
                "0",
                "-i",
                pathlist,
                "-c",
                "copy",
                str(path),
            ],
            stderr=subprocess.PIPE,
        )

        _, stderr = proc.communicate()
        assert proc.returncode == 0, stderr.decode()


def movie_wrapper(kwargs: dict[str, Any]) -> None:
    """
    Wrapper for movie to allow for multiprocessing.

    :param kwargs: Keyword arguments for :py:func:`movie`.
    """
    return movie(**kwargs)


def movie_progress_updater(
    progress: str, total: int, progress_dict: dict[str, int]
) -> None:
    """
    Update the progress bar for a movie.

    :param progress: Progress bar type.
    :param total: Total number of frames.
    :param progress_dict: Dictionary to store progress.
    """
    prog = PROGRESS_ITERS[progress]
    for i in prog(range(total)):
        while sum(progress_dict.values()) < i:
            time.sleep(0.1)


__all__ = [
    "imshow_defaults",
    "movie",
    "plot_frame",
]



from __future__ import annotations

import contextlib
import os
import re
from datetime import datetime
from typing import Any

import numpy as np
import pytest
import yaml

import pigreads as pig
from pigreads.schema.model import ModelDefinition


def test_str():
    pig.Models("marcotte2017dynamical")


def test_add():
    key = "marcotte2017dynamical"
    models = pig.Models(
        [
            (key, {}),
            (key, {"eps": 1}),
        ]
    )
    models.add(key)
    models.add(key, eps=2)
    assert models[0] == models[2]
    assert models[1].parameters["eps"] == 1
    assert models[3].parameters["eps"] == 2


def test_len():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    models.add(key)
    models.add(key)
    assert len(models) == 3


def test_Nv():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    assert models.Nv == 2


def test_Np():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    assert len(models[0].parameters) == 5


def test_set_param():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    models[0].parameters["eps"] *= 2


def test_del_param():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    with pytest.raises(NotImplementedError):
        del models[0].parameters["eps"]


def test_repr_params():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    repr(models[0].parameters)


def test_str_params():
    key = "marcotte2017dynamical"
    models = pig.Models()
    models.add(key)
    str(models[0].parameters)


def test_resting_states():
    models = pig.Models()
    models.add("gray1983autocatalytic")
    models.add("courtemanche1998ionic")
    inhom = np.array([[[0, 1], [2, 3]]], dtype=int)
    states = models.resting_states(inhom, Nframes=2)
    assert states.shape == (2, *inhom.shape, models.Nv)
    assert np.all(np.isnan(states[0, 0, 0, 0, :]))
    assert np.all(np.isnan(states[:, inhom == 0, :]))
    assert states[0, 0, 0, 1, 0] == 1
    assert states[0, 0, 0, 1, 1] == 0
    assert abs(-81.18 - states[0, 0, 1, 0, 0]) < 1e-6
    assert abs(11.17 - states[0, 0, 1, 0, 1]) < 1e-6
    assert states[0, 0, 1, 1, 0] == 1
    assert states[0, 0, 1, 1, 1] == 0
    assert np.all(np.isnan(states[1, inhom != 0, :]))


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_sim0d():
    z, y, x = np.mgrid[0:1, 0:3, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)

    for modelname in pig.Models.available:
        models = pig.Models(modelname)
        states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2).astype(
            np.float32
        )
        weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
        states[1] = models.run(inhom, weights, states[0], Nt=2, dt=1e-10)


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.skipif(
    "SKIP_DOUBLE" in os.environ,
    reason="Skipping tests with double precision via SKIP_DOUBLE",
)
def test_sim0d_double():
    z, y, x = np.mgrid[0:1, 0:3, 0:3]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)

    for modelname in pig.Models.available:
        models = pig.Models(modelname, double_precision=True)
        states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2).astype(
            np.float32
        )
        weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
        states[1] = models.run(inhom, weights, states[0], Nt=2, dt=1e-10)


def test_define():
    pig.Models.available["fitzhugh1961impulses"] = ModelDefinition(
        name="FitzHugh 1961 & Nagumo 1962",
        description="A 2D simplification of the Hodgkin-Huxley model.",
        dois=[
            "https://doi.org/10.1016/S0006-3495(61)86902-6",
            "https://doi.org/10.1109/JRPROC.1962.288235",
        ],
        variables={"u": 1.2, "v": -0.625},
        diffusivity={"u": 1.0},
        parameters={"a": 0.7, "b": 0.8, "c": 3.0, "z": 0.0},
        code="""
            *_new_u = u + dt * (v + u - u*u*u/3 + z + _diffuse_u);
            *_new_v = v + dt * (-(u - a + b*v)/c);
        """,
    )

    z, y, x = np.mgrid[0:1, 0:1, 0:1]
    dz, dy, dx = pig.deltas(z, y, x)
    inhom = np.ones_like(x, dtype=int)

    models = pig.Models("fitzhugh1961impulses")

    states: np.ndarray[Any, Any] = models.resting_states(inhom, Nframes=2)
    weights = models.weights(dz, dy, dx, inhom, diffusivity=0)
    states[1] = models.run(inhom, weights, states[0], Nt=100, dt=0.004)
    assert np.allclose(states[1], [1.1996392, -0.6249765])

    del pig.Models.available["fitzhugh1961impulses"]


def test_define_warning():
    cokey = "PYOPENCL_COMPILER_OUTPUT"

    if cokey in os.environ:
        del os.environ[cokey]

    for coval in [False, True]:
        if coval:
            os.environ[cokey] = "1"

        pig.Models.available["fitzhugh1961impulses"] = ModelDefinition(
            name="FitzHugh 1961 & Nagumo 1962",
            description="A 2D simplification of the Hodgkin-Huxley model.",
            dois=[
                "https://doi.org/10.1016/S0006-3495(61)86902-6",
                "https://doi.org/10.1109/JRPROC.1962.288235",
            ],
            variables={"u": 1.2, "v": -0.625},
            diffusivity={"u": 1.0},
            parameters={"a": 0.7, "b": 0.8, "c": 3.0, "z": 0.0},
            code=f"""
                *_new_u = u + dt * (v + u - u*u*u/3 + z + _diffuse_u);
                *_new_v = v + dt * (-(u - a + b*v)/c);
                #warning "Test warning! {datetime.now()}"
            """,
        )

        with contextlib.suppress(UserWarning):
            pig.Models("fitzhugh1961impulses")

        if coval:
            del os.environ[cokey]

    del pig.Models.available["fitzhugh1961impulses"]


def test_block_size():
    models = pig.Models()
    assert models.block_size == (1, 8, 8)
    models.block_size = (8, 8, 8)
    assert models.block_size == (8, 8, 8)


def test_unused():
    unused: dict[str, dict[str, dict[str, float]]] = {}
    for name, model in pig.Models.available.items():
        unused[name] = {
            "variables": {},
            "diffusivity": {},
            "parameters": {},
            "updates": {},
        }

        for parameter, value in model.parameters.items():
            if not re.search(rf"\b{parameter}\b", model.code):
                unused[name]["parameters"][parameter] = value

        for var, value in model.diffusivity.items():
            if not re.search(rf"\b_diffuse_{var}\b", model.code):
                unused[name]["diffusivity"][var] = value

        for var, value in model.variables.items():
            if not re.search(rf"^\s*\*_new_{var}\s*=", model.code, flags=re.MULTILINE):
                unused[name]["variables"][var] = value

            new = f"_new_{var}"
            if not re.search(rf"^\s*\*{new}\s*=", model.code, flags=re.MULTILINE):
                unused[name]["updates"][new] = np.nan

        unused[name] = {k: v for k, v in unused[name].items() if len(v) > 0}
        if len(unused[name]) == 0:
            del unused[name]

    assert len(unused) == 0, "Unused symbols:\n" + yaml.safe_dump(
        unused, sort_keys=False
    )
